/*
  Extension     FileTabWidth
  Author        Carlo Hogeveen
  Website       eCarlo.nl/tse
  Compatibility TSE v4 upwards, all variants
  Version       v1   22 May 2025


  NOTICE
    This extension is formally still in developement, but might be ready.
    Let me know if you have any criticisms, comments or wishes.

  This extension
  - sets TSE's TabWidth to each file's tab width,
  - on opening a file optionally detabs it with this TabWidth,
  - or detabs a file with it when you tell it to,
  - and optionally shows each file's tab width as a status.

  A file's tab width is based on the file's indentations when opened.
  A configurable default tab width is used when no file tab width can be
  determined.

  The extension is incompatible with TSE's "Detab files when loading" option.

  However, it now has its own "Detab on file open" configuration option.
  The difference and improvement is, that TSE used to detab a file with your
  default tab width, while this extension will detab a file with the file's
  tab width!

  The extension is incompatible with TSE's "Zero-Based Tab Stops" option.
  Research says it its use is extremely rare.
  Let me know if you do have a good use-case for them.


  INSTALLATION

    Turn off TSE's configuration option "Detab files when loading" and
    "Save Current Settings".
    It is incompatible with FileTabWidth, because it would make TSE detab an
    opened file using the tab width of the previous current file.
    FileTabWith comes with is own detab functionalities instead. Read on.

    Copy this file to TSE's "mac" directory, and compile it there, for example
    by opening it there in TSE and applying the Macro Compile menu.

    Add this extension's name "FileTabWidth" to TSE's Macro AutoLoad List menu.

    Optionally Macro Execute "FileTabWidth" to configure it:
      Show tab width as a status
        Immediately see each current file's tab width as "Tab <number>".
        Only works if at least v1.2 of the Status extension is installed:
          https://ecarlo.nl/tse/ToolsAndExtensions.html#Status
      Detab on file open (Yes/No)
        Detabs each file when it is opened. (*1)
      Default tab width
        The extension automatically takes over TSE's default tab width.
        You will no longer be able to set a default tab width in TSE, because
        TSE's tab width now continously adjusts to the current file's tab width.

    Note:
      If you are a user of the SynCase extension, then make sure you have at
      least version 7.1 (6 Oct 2022) installed.
      An older version had a flaw that caused FileTabWidth to show tab width 1
      just after opening a .s macro source file and moving the cursor.


  USAGE

    Normally you will have to do nothing.

    TSE's tab width automatically gets set to the current file's tab width.
    If a file has no determinable tab width, the tab width is set to the
    default tab width.

    The extension's menu has options for the current file:
    - Tab Width
      Determines the current file's tab width setting for the duration of
      the TSE session.
    - Detab ...
      Detabs the current file. (*1)
      This option is only available if the current file has tab characters.

    If the "Show tab width as a status" option is "On", then you can click on
    the "Tab" status to open FileTabWidth's menu.


  (*1)
    "Detabbing" a file:

    In the file each tab character is replaced with the amount of spaces
    that match the file's tab width setting and TSE's "Zero-Based Tab Stops"
    setting.

    A test case are the lines 149 - 152 of "mac\decomp.s", which use tab
    characters for indentation.

    If there were tab characters, then the file is marked as changed.
    The past teaches that this is a controversial choice.
    An advantage is, that the user is informed that the file is changed.
    A disadvantage is, that when you open (possibly a lot of) files just to
    view them and/or search their content, then you cannot easily close them
    without having to confirm for each changed file whether you want to save it
    or not.
    I have negated this disadvantage by providing new versions of TSE's "Exit"
    and "find&do" functions, that offer two extra buttons, "Yes to all"
    and "No to all".


  TODO
    MUST
    SHOULD
    - ?
    COULD
    - Add a menu option to define a key to open the menu.
    - The menu already lets us change the current file's tab width for the
      duration of the TSE session.
      This does not change the file itself, but might be useful in case of a
      file tab width that was wrongly determined automatically.
      Extend this capability to future TSE sessions?
    WONT


  HISTORY

  v0.0.0.1    18 May 2025
    Initial development release.
    Can set a file's tab width on opening it, and remember it between switching
    files.

  v0.0.0.2    18 May 2025 (2)
    FileTabWidth's menu now lets you change the current file's tab width for
    the duration of the TSE session.

  v0.0.0.3    18 May 2025 (3)
    Added a menu option to detab the current file.

  v0.0.0.4    19 May 2025
    Added a menu option to toggle whether a file should be detabbed on opening.

  v0.0.0.5    20 May 2025
    An apparent bug was caused by find&do disabling hooks when it loaded files,
    which blocked FileTabWidth from setting a file's tab width and detabbing a
    file, and blocked Unicode files from being converted when opened.
    In effect find&do was side-loading "raw" files.
    I have a fix for find&do and will be submitting it shortly.
    As long as find&do's bug is not fixed, FileTabWidth will now use the
    less-wrong default tab width for a file AND give a warning.

  v0.0.0.6    21 May 2025
    Research says zero-based tab stops are very rare.
    FileTabWidth will not support them,
    and now aborts if TSE setting "Zero-Based Tab Stops" is On.
    Let me know if you do have a good use-case for them.

  v0.0.0.7    21 May 2025 (2)
    Added a menu option to show the tab width as a status.
    Usage requires at least v1.2 of the Status extension to be installed.

  v0.0.0.8    21 May 2025 (3)
    Only documentation improvemonets to prepare it for the v1 release.

  v0.0.0.9    22 May 2025
    Documented the major bug it appeared to have in Linux:
    I found the bug, that the first opened file initially gets the correct tab
    width, then when browsed gets tab width 1, and when switching between files
    gets the correct tab width again.

  v0.0.0.10   22 May 2025 (2)
    The "Linux bug" was caused by me having an old and apparently flawed
    version of the SynCase macro installed in my Linux TSE.
    Updating SynCase to its latest version 7.1 (6 Oct 2022) fixed the bug.

  v1          22 May 2025
    Initial non-development release.
    Did add some fixes to actually make it backwards compatible to TSE v4.

*/





/*
  T E C H N I C A L   B A C K G R O U N D   I N F O


  NOT USING TSE'S "DETAB FILES WHEN LOADING" SETTING

    Tests say that in theory each file's TabWidth could be set
    before DetabOnLoad is executed by hooking _ON_FILE_LOAD_.
    _ON_FIRST_EDIT_ would be too late.
    A significant penalty would be, that the _ON_FILE_LOAD_ hook would need to
    preload the file to determine its tab size, meaning each file would be
    loaded twice.
    Based on this and some other penalties I decided against this algorithm
    in favor of one that determines the tab size later and optionally does its
    own detabbing.

*/





//  Start of get_tab_size.inc ...



/*
  Include-file    get_tab_size.inc
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE v4 upwards, all variants
  Version         v2.0.1   18 May 2025


  For macro programmers this include-file implements the function

    integer get_tab_size()

  It returns the current buffer's best tab size, or 0 if none could be
  determined.

  Here tab size means how much space is consistently used to increase lines'
  indentations.

  See the "ALGORITHM" topic in the "TECHNICAL BACKGROUND INFORMATION" section
  if you really want to know some of the dirty details.

  This include-file comes with a tool get_tab_size_test to test and demonstrate
  its use.

  Terminology:
    "TabWidth"
      is the name of a TSE setting used for editing all buffers.
    "Tab size"
      can either refer to an editor setting in general, or to the editor
      setting that apparently was used to edit a specific file.
      In the latter case we can say that the file has a tab size.


  DISCLAIMER

    For a tiny minority of buffers no sensible tab size can be determined,
    for example for empty buffers, or for buffers that use different tab sizes
    in different parts of the buffer, or for buffers that otherwise do not use
    consistent tab sizes, or for buffers that only use tab characters for
    indentation.
    And in an even tinier minority of cases get_tab_size may simply fail to
    determine what the human intended.


  USAGE

  - Either include get_tab_size.inc in your macro with an #include statementment
    like
      #include ["get_tab_size.inc"]
    or copy/paste get_tab_size.inc's source code as a whole into your macro.

    I prefer the copy/paste method. Because I distribute macros, I do not want
    to impose on users of my macros the need to do dependency management of the
    versions of macros and the versions of something as technical as .inc files.

    An great example of this is, that this specific version of get_tab_size.inc
    includes a copy/pasted copy of a specific version of int_array.inc's source
    code.

  - For debugging get_tab_size() you can put
      #define GET_TAB_SIZE_DEBUG 1
    before get_tab_size.inc's code.

    This will create a new buffer with get_tab_size()'s intermediate results.


  HISTORY

  v1      24 Aug 2021
    Initial release.
    Only worked for files without tab characters.


  v2      16 May 2025
    Now also works for files with tab characters.

    Changes in v2 that break backwards compatibility with v1:
    - If no tab size can be determined, then now 0 is returned instead of 1.
      For example for empty files, files with no indentation, or files that
      only use Tab characters for indentation.
    - get_tab_size.inc is now an include file:
      It is no longer executable as a macro, and no longer has parameters.
    - It can no longer create a debug report.


  v2.0.1  18 May 2025
    Fixed "ERROR 6" warning.

*/



/*
  T E C H N I C A L    B A C K G R O U N D   I N F O R M A T I O N


  ALGORITHM

  To determine the current buffer's tab size the tool counts the relative
  increase in indentation between consecutive lines. Tests show that this
  simple method works better than other simple methods.

  Take this example, that only has tabbed indentation.
  The example uses the ^ character to indicate the start of a line:
    ^if a
    ^  if b
    ^    x = 1
    ^  endif
    ^endif
  The 5 lines have absolute indentations 0, 2, 4, 2 and 0.
  The algorithm only looks at relative increases, so between line 1 and 2 and
  between line 2 and 3, giving tab sizes 2 and 2.

  The algorithm is even a bit smarter than that.

  The following example has tabbed and alignment indentation:
    ^if a
    ^  if (   b >= 1
    ^     and b <= 3 )
    ^    x = 1
    ^  endif
    ^endif
  Line 3 does not have tabbed indentation but alignment indentation!
  The 5 lines have absolute indentations 0, 2, 5, 4, 2 and 0 respectively.
  Between lines 2 and 3 the indentation increases with 3 to a total of 5, but
  because 5 is not a multiple of 3, this cannot be tabbed intdentation, and the
  line is ignored.
  The algorithm therefore sees an indentation increase of 2 between lines 2
  and 4, and counts that as a possible tab size of 2. Allrighty!

  Of course there are sequences of lines that make the algorithm fail, but
  tests in practice so far show, that the determined tab size for the whole
  buffer is always defendable, even though in a few worst cases that might be
  because there is no one sensible tab size for a particular buffer.


  TAB SEPARATED VALUES

    Tab-separated value (TSV) files have nothing to to with indentation.

    Like the frequntly occurring comma-separated value files and
    semi-colon-separated value files, in theory tab-separated value files can
    also exist, and I wondered whether to handle them as an exceptional case.

    I scanned my computer, which has software from many different sources, and
    found that TSV files do exist in the wild, but are very rare.

    So rare, that I decided not to handle them.

    Note:
      I found that TSV files usually does not have extension .tsv.

    Note:
      Here is a related, off-putting technical tidbit I encountered.
      If the last field of a line in a TSV file is empty,
      and TSE's RemoveTrailingWhite setting is ON,
      then TSE uses non-intuitive rules for removing trailing tab characters.
      For eample:
        On the first line a trailing tab character is removed when the file is
        loaded.
        A trailing tab character is removed after it is searched and found,
        so you will find it only once.
      This behavior is not necessarily wrong, but might be unexpected.

*/





// Start of compatibility restrictions and mitigations ...


#ifdef LINUX
  #define WIN32 FALSE
#endif

#ifndef WIN32
  16-bit versions of TSE are not supported. You need at least TSE v4.
#endif

#ifndef EDITOR_VERSION
  Editor Version is older than TSE v3. You need at least TSE v4.
#endif

#if EDITOR_VERSION < 4000h
  Editor Version is older than TSE v4. You need at least TSE v4.
#endif

#ifndef INTERNAL_VERSION
  #define INTERNAL_VERSION 0
#endif


#if EDITOR_VERSION < 4400h
  /*
    StrReplace() 1.0

    If you have TSE Pro 4.0 or 4.2, then this proc almost completely implements
    the built-in StrReplace() function of TSE Pro 4.4.
    The StrReplace() function replaces a string (pattern) inside a string.
    It works for strings like the Replace() function does for files, so read
    the Help for the Replace() function for the usage of the options, but apply
    these differences while reading:
    - Where Replace() refers to "file" and "line", StrReplace() refers to
      "string".
    - The options "g" ("global", meaning "from the start of the string")
      and "n" ("no questions", meaning "do not ask for confirmation on
      replacements") are implicitly always active, and can therefore be omitted.
    Notable differences between the procedure below with TSE 4.4's built-in
    function are, that here the fourth parameter "options" is mandatory
    and that the fifth parameter "start position" does not exist.
  */
  integer strreplace_id = 0
  string proc StrReplace(string needle, string haystack, string replacer, string options)
    integer i                      = 0
    integer org_id                 = GetBufferId()
    string  result  [MAXSTRINGLEN] = haystack
    string  validated_options [20] = 'gn'
    for i = 1 to Length(options)
      if (Lower(SubStr(options, i, 1)) in '0'..'9', 'b', 'i','w', 'x', '^', '$')
        validated_options = validated_options + SubStr(options, i, 1)
      endif
    endfor
    if strreplace_id == 0
      strreplace_id = CreateTempBuffer()
    else
      GotoBufferId(strreplace_id)
      EmptyBuffer()
    endif
    InsertText(haystack, _INSERT_)
    lReplace(needle, replacer, validated_options)
    result = GetText(1, CurrLineLen())
    GotoBufferId(org_id)
    return(result)
  end StrReplace
#endif


/*
  compare_versions()  v2.0

  This proc compares two version strings version1 and version2, and returns
  -1, 0 or 1 if version1 is smaller, equal, or greater than/to version2.

  For the comparison a version string is split into parts:
  - Explicitly by separating parts by a period.
  - Implicitly:
    - Any uninterrupted sequence of digits is a "number part".
    - Any uninterrupted sequence of other characters is a "string part".

  Spaces are mostly ignored. They are only significant:
  - Between two digits they signify that the digits belong to different parts.
  - Between two "other characters" they belong to the same string part.

  If the first version part is a single "v" or "V" then it is ignored.

  Two version strings are compared by comparing their respective version parts
  from left to right.

  Two number parts are compared numerically, e.g: "1" < "2" < "11" < "012".

  Any other combination of version parts is case-insensitively compared as
  strings, e.g: "12" < "one" < "three" < "two", or "a" < "B" < "c" < "d".

  Examples: See the included unit tests further on.

  v2.0
    Out in the wild there is an unversioned version of compare_versions(),
    that is more restricted in what version formats it can recognize,
    therefore here versioning of compare_versions() starts at v2.0.
*/

// compare_versions_standardize() is a helper proc for compare_versions().

string proc compare_versions_standardize(string p_version)
  integer char_nr                  = 0
  string  n_version [MAXSTRINGLEN] = Trim(p_version)

  // Replace any spaces between digits by one period. Needs two StrReplace()s.
  n_version = StrReplace('{[0-9]} #{[0-9]}', n_version, '\1.\2', 'x')
  n_version = StrReplace('{[0-9]} #{[0-9]}', n_version, '\1.\2', 'x')

  // Remove any spaces before and after digits.
  n_version = StrReplace(' #{[0-9]}', n_version, '\1', 'x')
  n_version = StrReplace('{[0-9]} #', n_version, '\1', 'x')

  // Remove any spaces before and after periods.
  n_version = StrReplace(' #{\.}', n_version, '\1', 'x')
  n_version = StrReplace('{\.} #', n_version, '\1', 'x')

  // Separate version parts by periods if they aren't yet.
  char_nr = 1
  while char_nr < Length(n_version)
    case n_version[char_nr:1]
      when '.'
        NoOp()
      when '0' .. '9'
        if not (n_version[char_nr+1:1] in '0' .. '9', '.')
          n_version = n_version[1:char_nr] + '.' + n_version[char_nr+1:MAXSTRINGLEN]
        endif
      otherwise
        if (n_version[char_nr+1:1] in '0' .. '9')
          n_version = n_version[1:char_nr] + '.' + n_version[char_nr+1:MAXSTRINGLEN]
        endif
    endcase
    char_nr = char_nr + 1
  endwhile
  // Remove a leading 'v' if it is by itself, i.e not part of a non-numeric string.
  if  (n_version[1:2] in 'v.', 'V.')
    n_version = n_version[3:MAXSTRINGLEN]
  endif
  return(n_version)
end compare_versions_standardize

integer proc compare_versions(string version1, string version2)
  integer result                 = 0
  string  v1_part [MAXSTRINGLEN] = ''
  string  v1_str  [MAXSTRINGLEN] = ''
  string  v2_part [MAXSTRINGLEN] = ''
  string  v2_str  [MAXSTRINGLEN] = ''
  integer v_num_parts            = 0
  integer v_part_nr              = 0
  v1_str      = compare_versions_standardize(version1)
  v2_str      = compare_versions_standardize(version2)
  v_num_parts = Max(NumTokens(v1_str, '.'), NumTokens(v2_str, '.'))
  repeat
    v_part_nr = v_part_nr + 1
    v1_part   = Trim(GetToken(v1_str, '.', v_part_nr))
    v2_part   = Trim(GetToken(v2_str, '.', v_part_nr))
    if  v1_part == ''
    and isDigit(v2_part)
      v1_part = '0'
    endif
    if v2_part == ''
    and isDigit(v1_part)
      v2_part = '0'
    endif
    if  isDigit(v1_part)
    and isDigit(v2_part)
      if     Val(v1_part) < Val(v2_part)
        result = -1
      elseif Val(v1_part) > Val(v2_part)
        result =  1
      endif
    else
      result = CmpiStr(v1_part, v2_part)
    endif
  until result    <> 0
     or v_part_nr >= v_num_parts
  return(result)
end compare_versions


// End of compatibility restrictions and mitigations.





//  Start of int_array implementation ...

/*
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Version         v1   14 May 2025
  Compatibility   TSE v4 upwards, all variants

  Only use the following functions to manipulate integer arrays:
    integer proc int_array_create(string array_name)
    integer proc int_array_delete(string array_name)
    integer proc int_array_set(string array_name, integer index, integer value)
    integer proc int_array_get(string array_name, integer index)

  int_array_get() returns the value at the index, which defaults to 0.
  All other functions return TRUE or FALSE if they went ok or not.

  An array is intended to only be used in the macro it is created in.
  An array name only needs to be unique within the macro it is created in.
  You can create as many arrays as TSE has free memory and bufferids.
  Creating an array will use up a bufffer id, which limits you to MAXINT
  arrays minus how many bufferids other TSE programs use.
  Array elements can have an index from MININT to MAXINT.
  When an array is created, all array elements exist with value 0.
  Array elements with value 0 do not use memory.
  Setting an array element from not 0 to 0 frees up its memory usage.
  Deleting an array will free up all its memory usage at once.


  NOTE
    I maintain int_array's source as a separate include-file,
    and as an independent and versioned source of truth.

    But for actual use I copy int_array's source as a whole into each macro
    that needs it, so that each such a macro can be tested and distributed
    without int_arrays's version being an outside dependency.

*/


// Do not use or call these int_array_internal_... variables and functions!

integer int_array_internal_id                    = 0
string  int_array_internal_prefix [MAXSTRINGLEN] = 'IntArray'


proc int_array_internal_context()
  integer org_id = 0

  if int_array_internal_id == 0
    int_array_internal_prefix = SplitPath(CurrMacroFilename(), _NAME_) + ':IntArray'
    org_id                    = GetBufferId()
    int_array_internal_id     = CreateTempBuffer()
    ChangeCurrFilename(int_array_internal_prefix, _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
    int_array_internal_prefix = int_array_internal_prefix + ':'
    GotoBufferId(org_id)
  endif
end int_array_internal_context


proc int_array_internal_warn(string array_name, string msg)
  if (Query(MsgLevel) in _ALL_MESSAGES_, _WARNINGS_ONLY_)
    if Query(Beep)
      Alarm()
    endif

    MsgBox(int_array_internal_prefix + array_name,
           iif(Val(msg),
               'ERROR ' + msg + '.',
               'ERROR: ' + msg))
  endif
end int_array_internal_warn

// End of int_array_internal_... variables and functions.


// Start of externally callable int_array functions:

integer proc int_array_create(string array_name)
  integer array_id = 0
  integer ok       = FALSE
  integer org_id   = GetBufferId()

  int_array_internal_context()

  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)
  if array_id
    int_array_internal_warn(array_name, '1')
  else
    array_id = CreateTempBuffer()

    if array_id
      ChangeCurrFilename(int_array_internal_prefix + array_name,
                         _DONT_PROMPT_|_DONT_EXPAND_|_OVERWRITE_)
      if SetBufferInt(int_array_internal_prefix + array_name,
                      array_id,
                      int_array_internal_id)
        ok = TRUE
      else
        int_array_internal_warn(array_name, '2')
      endif
    else
      int_array_internal_warn(array_name, '3')
    endif

    GotoBufferId(org_id)
  endif

  return(ok)
end int_array_create


integer proc int_array_delete(string array_name)
  integer array_id = 0
  integer ok       = FALSE

  int_array_internal_context()
  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)
  if array_id
    ok = TRUE

    if not AbandonFile(array_id)
      int_array_internal_warn(array_name, '4')
      ok = FALSE
    endif

    if not DelBufferVar(int_array_internal_prefix + array_name,
                        int_array_internal_id)
      int_array_internal_warn(array_name, '5')
      ok = FALSE
    endif
  else
    int_array_internal_warn(array_name, '6')
  endif

  return(ok)
end int_array_delete


integer proc int_array_set(string array_name, integer index, integer value)
  integer array_id = 0
  integer ok       = FALSE

  int_array_internal_context()
  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)

  if array_id
    if value
      if SetBufferInt(int_array_internal_prefix + array_name + Str(index),
                      value,
                      array_id)
        ok = TRUE
      else
        int_array_internal_warn(array_name, '7')
      endif
    else
      DelBufferVar(int_array_internal_prefix + array_name + Str(index),
                   array_id)
    endif
  else
    int_array_internal_warn(array_name, '8')
  endif

  return(ok)
end int_array_set


integer proc int_array_get(string array_name, integer index)
  integer array_id = 0
  integer value    = 0

  int_array_internal_context()
  array_id = GetBufferInt(int_array_internal_prefix + array_name,
                          int_array_internal_id)

  if array_id
    value = GetBufferInt(int_array_internal_prefix + array_name + Str(index),
                         array_id)
  endif

  return(value)
end int_array_get

// End of int_array implementation.





// Get tab size implementation


#ifdef GET_TAB_SIZE_DEBUG
  proc get_tab_size_debug_report(integer best_tab_size,
                                 integer max_tab_size)
    string  curr_file_name [MAXSTRINGLEN] = CurrFilename()
    integer first_log_line                = 0
    integer last_log_line                 = 0
    integer tab_size                      = 0

    NewFile()
    SetUndoOff()
    AddLine('')
    AddLine(curr_file_name)
    AddLine('')
    AddLine(Format('Tab Size':10, 'Occurs':10))
    first_log_line = CurrLine() + 1

    for tab_size = 1 to max_tab_size
      if int_array_get('tab_size_count', tab_size)
        AddLine(Format(tab_size:10, int_array_get('tab_size_count', tab_size):10))
      endif
    endfor

    last_log_line = CurrLine()

    if first_log_line < last_log_line
      PushBlock()
      MarkLine(first_log_line, last_log_line)
      Sort()
      PopBlock()
    endif

    EndLine()
    AddLine('')
    AddLine(Format('Best tab size:'; best_tab_size))
    AddLine('')
    SetUndoOn()
    BegFile()
    UpdateDisplay()
  end get_tab_size_debug_report
#endif

/*
  This proc counts each indentation increase as an occurrence of a tab size
  if the new indentation is a multiple of the increase,
  and it also counts it as an occurrence of the same tab size if such an
  indentation is continued on a consecutive line.

*/
integer proc get_tab_size()
  integer best_tab_size        = 0
  integer best_tab_size_count  = 0
  integer indentation          = 0
  integer indentation_increase = 0
  integer max_tab_size         = 0
  integer prev_indentation     = 1
  integer tab_size             = 0
  string  text  [MAXSTRINGLEN] = ''

  if NumLines()
    int_array_create('tab_size_count')
    PushLocation()
    BegFile()

    repeat
      text = GetText(1, MAXSTRINGLEN)

      if not Pos(Chr(9), text)
      or     Pos(Chr(9), text) > PosFirstNonWhite()
        indentation = PosFirstNonWhite()

        if indentation
          indentation_increase = indentation - prev_indentation

          if indentation_increase >= 0
            if indentation_increase             // > 0
              if (indentation - 1) mod indentation_increase
                // Invalid tab size: negative dummy value to ignore this line.
                indentation_increase = -1
              else
                tab_size = indentation_increase // New tab size.
              endif
            else
              // Same indentation: positive dummy value to count last tab size.
              indentation_increase = +1
            endif
          else
            tab_size = 0 // Negative increase: ignore line and restart scratch.
          endif

          if  tab_size
          and indentation_increase >= 0
            int_array_set('tab_size_count',
                          tab_size,
                          int_array_get('tab_size_count', tab_size) + 1)
          endif

          prev_indentation = indentation

          if tab_size > max_tab_size
            max_tab_size = tab_size
          endif
        endif
      endif
    until not Down()

    PopLocation()

    for tab_size = 1 to max_tab_size
      if int_array_get('tab_size_count', tab_size) > best_tab_size_count
        best_tab_size_count = int_array_get('tab_size_count', tab_size)
        best_tab_size       = tab_size
      endif
    endfor

    #ifdef GET_TAB_SIZE_DEBUG
      get_tab_size_debug_report(best_tab_size, max_tab_size)
    #endif

    int_array_delete('tab_size_count')
  endif

  return(best_tab_size)
end get_tab_size


//  End of get_tab_size.inc.







//  Constants and semi-constants
#define MAX_IDLE_TIME                           36
string  MACRO_NAME             [MAXSTRINGLEN] = ''
string  SHOW_STATUS_CALL       [MAXSTRINGLEN] = ''
string  STATUS_MACRO_NAME      [MAXSTRINGLEN] = 'Status'
string  STATUS_MACRO_REQUIRED_VERSION     [3] = '1.2'
string  VARNAME_FILE_TAB_WIDTH [MAXSTRINGLEN] = ''


//  Global variables
integer default_tab_width   = 0
integer detab_on_first_edit = FALSE
integer idle_timer          = 0
integer macro_is_ok         = TRUE
integer show_status         = FALSE


proc warning(string msg)
  if Query(Beep)
    Alarm()
  endif

  MsgBox(MACRO_NAME, msg)
end warning


integer proc macro_ok()
  #if EDITOR_VERSION >= 4200h
    if  macro_is_ok
    and Query(DetabOnLoad)
      warning(MACRO_NAME + ' aborted.' + Chr(13) +
              'It is incompatible with TSE setting "Detab files when loading" being "On".')
      macro_is_ok = FALSE
      PurgeMacro(MACRO_NAME)
    endif
  #endif

  if  macro_is_ok
  and Query(ZeroBasedTab)
    warning(MACRO_NAME + ' aborted.' + Chr(13) +
            'It is incompatible with TSE setting "Zero-Based Tab Stops" being "On".')
    macro_is_ok = FALSE
    PurgeMacro(MACRO_NAME)
  endif

  return(macro_is_ok)
end macro_ok


/*
  An ExecMacro that eventually stops this macro if the Status macro is not
  installed, thereby avoiding an infinite loop.
*/
proc exec_macro(string macro_cmd_line)
  string  status_macro_current_version [MAXSTRINGLEN] = ''
  string  extra_version_text           [MAXSTRINGLEN] = ''
  integer ok                                          = TRUE

  ok = ExecMacro(macro_cmd_line)

  if ok
    // New way: Lower-case macro name in global variable name.
    status_macro_current_version = GetGlobalStr(Lower(STATUS_MACRO_NAME) + ':Version')

    if status_macro_current_version == ''
      // Old way: Cased macro name in global variable name.
      status_macro_current_version = GetGlobalStr(STATUS_MACRO_NAME + ':Version')
    endif

    if status_macro_current_version == ''
      ok = FALSE
    elseif compare_versions(status_macro_current_version,
                            STATUS_MACRO_REQUIRED_VERSION) == -1
      ok                 = FALSE
      extra_version_text = 'at least version ' +
                           STATUS_MACRO_REQUIRED_VERSION + ' of '
    endif
  endif

  if not ok
    warning(MACRO_NAME + ' aborted: It needs ' + extra_version_text
            + 'the "' + STATUS_MACRO_NAME + '" macro to be installed!')
    macro_is_ok = FALSE
    PurgeMacro(MACRO_NAME)
  endif

end exec_macro


integer proc write_profile_int(string  section_name,
                               string  item_name,
                               integer item_value)
  integer ok = WriteProfileInt(section_name,
                               item_name,
                               item_value)
  if not ok
    if Query(Beep)
      Alarm()
    endif
    MsgBox(MACRO_NAME, 'Could not write item "' + item_name +
                       '" to configuration file "tse.ini".')
  endif

  return(ok)
end write_profile_int


proc detab_current_file()
  integer old_Insert = Set(Insert, ON)
  integer tab_width  = Query(TabWidth)

  PushPosition()

  if lFind('\d009', 'gx')
    repeat
      DelChar()
      InsertText(Format('': tab_width - CurrPos() mod tab_width))
    until not lFind('\d009', 'x')
  endif

  PopPosition()
  Set(Insert, old_Insert)
end detab_current_file


proc on_first_edit()
  integer file_tab_width = 0

  idle_timer = MAX_IDLE_TIME

  if macro_ok()
    file_tab_width = get_tab_size()

    if file_tab_width == 0
      file_tab_width = default_tab_width
    endif

    Set(TabWidth, file_tab_width)
    SetBufferInt(VARNAME_FILE_TAB_WIDTH, file_tab_width)

    if detab_on_first_edit
      detab_current_file()
    endif
  endif
end on_first_edit


proc on_changing_files()
  #if INTERNAL_VERSION > 0
    integer use_home_path = _USE_HOME_PATH_
  #else
    integer use_home_path = MAXSTRINGLEN
  #endif

  idle_timer = MAX_IDLE_TIME

  if macro_ok()
    if GetBufferInt(VARNAME_FILE_TAB_WIDTH)
      Set(TabWidth, GetBufferInt(VARNAME_FILE_TAB_WIDTH))
    else
      UpdateDisplay(_ALL_WINDOWS_REFRESH_)
      MsgBox(MACRO_NAME,
             "The file has no tab width, which might be caused by loading it with an old"
             + Chr(13)
             + "find&do version. The file is given the less wrong default tab width instead."
             + Chr(13)
             + SqueezePath(CurrMacroFilename(), use_home_path))
      Set(TabWidth, default_tab_width)
    endif
  endif
end on_changing_files


proc idle()
  idle_timer = idle_timer + 1

  if idle_timer >= MAX_IDLE_TIME
    idle_timer = 0
    exec_macro(SHOW_STATUS_CALL + Str(Query(TabWidth)))
  endif
end idle


integer proc read_num(integer n)
  string s [3] = Str(n)

  return (iif(ReadNumeric(s), Val(s), n))
end read_num


proc set_default_tab_width()
  integer new_value = read_num(default_tab_width)

  if new_value in 1 .. 255
    default_tab_width = new_value
    write_profile_int(MACRO_NAME + ':Configuration', 'DefaultTabWidth',
                      default_tab_width)
  else
    warning('Value (' + Str(new_value) + ') out of range (1 .. 255).')
  endif
end set_default_tab_width


proc set_current_file_tab_width()
  integer new_value = read_num(Query(TabWidth))

  if new_value in 1 .. 255
    Set(TabWidth, new_value)
    SetBufferInt(VARNAME_FILE_TAB_WIDTH, new_value)
  else
    warning('Value (' + Str(new_value) + ') out of range (1 .. 255).')
  endif
end set_current_file_tab_width


integer proc get_detab_flag()
  integer flag = _MF_DISABLED_|_MF_GRAYED_

  PushLocation()

  if lFind('\d009', 'gx')
    flag = _MF_ENABLED_|_MF_CLOSE_ALL_BEFORE_
    PopLocation()
  else
    KillLocation()
  endif

  return(flag)
end get_detab_flag


string proc get_detab_on_first_edit()
  return(iif(detab_on_first_edit, 'Yes', 'No'))
end get_detab_on_first_edit


proc toggle_detab_on_first_edit()
  detab_on_first_edit = not detab_on_first_edit

  write_profile_int(MACRO_NAME + ':Configuration', 'DetabOnFirstEdit',
                    detab_on_first_edit)
end toggle_detab_on_first_edit


integer proc get_show_status_flag()
  integer flag = _MF_DISABLED_|_MF_GRAYED_

  if FileExists(SearchPath('Status.mac', Query(TSEPath), 'mac'))
    flag = _MF_ENABLED_|_MF_DONT_CLOSE_
  endif

  return(flag)
end get_show_status_flag


string proc get_show_status()
  return(iif(show_status, 'Yes', 'No'))
end get_show_status


proc toggle_show_status()
  show_status = not show_status

  if show_status
    Hook(_IDLE_, idle)
  else
    UnHook(idle)
  endif

  write_profile_int(MACRO_NAME + ':Configuration', 'ShowStatus', show_status)
end toggle_show_status


menu main_menu()
  title = 'FileTabWidth configuration menu'

  'Current File',, _MF_DIVIDE_

  '&Tab width'
    [Str(Query(TabWidth)): 3],
    set_current_file_tab_width(),
    _MF_ENABLED_|_MF_DONT_CLOSE_,
    "Changes the current file's tab width for the duration of the TSE session."

  '&Detab ...',
    detab_current_file(),
    get_detab_flag(),
    "Replace each tab character with spaces based on the current file's tab width."

  'General settings',, _MF_DIVIDE_

  'Show tab width as a &status'
    [get_show_status():3],
    toggle_show_status(),
    get_show_status_flag(),
    "Toggle whether to show tab width as a status. Requires the Status extension."

  'Detab on file &open'
    [get_detab_on_first_edit():3],
    toggle_detab_on_first_edit(),
    _MF_ENABLED_|_MF_DONT_CLOSE_,
    'Toggle whether to remove tab characters from files when they are opened.'

  'Default tab &width'
    [Str(default_tab_width): 3],
    set_default_tab_width(),
    _MF_ENABLED_|_MF_DONT_CLOSE_,
    'Set the default tab width for files with an undeterminable tab width.'

end main_menu


proc WhenLoaded()
  MACRO_NAME             = SplitPath(CurrMacroFilename(), _NAME_)
  SHOW_STATUS_CALL       = STATUS_MACRO_NAME + ' ' + MACRO_NAME + ':TabWidth,callback Tab '
  VARNAME_FILE_TAB_WIDTH = MACRO_NAME + ':FileTabWidth'

  if macro_ok()
    default_tab_width = GetProfileInt(MACRO_NAME + ':Configuration',
                                      'DefaultTabWidth', 0)
    if default_tab_width == 0
      default_tab_width = Query(TabWidth)
      write_profile_int(MACRO_NAME + ':Configuration', 'DefaultTabWidth',
                        default_tab_width)
    endif

    detab_on_first_edit = GetProfileInt(MACRO_NAME + ':Configuration',
                                        'DetabOnFirstEdit', FALSE)

    if get_show_status_flag() == _MF_ENABLED_|_MF_DONT_CLOSE_
      show_status = GetProfileInt(MACRO_NAME + ':Configuration', 'ShowStatus',
                                  FALSE)
      if show_status
        Hook(_IDLE_, idle)
      endif
    else
      show_status = FALSE
    endif

    Hook(_ON_FIRST_EDIT_    , on_first_edit)
    Hook(_ON_CHANGING_FILES_, on_changing_files)
  endif
end WhenLoaded


proc Main()
  if macro_ok()
    main_menu()
  endif
end Main

